/*
 * Copyright 2010 Cyril Crassin.
 * v1.5
 *
 * This sample shows how dynamic template parameters can be emulated.
 * Cf. http://blog.icare3d.org
*/


// includes, system
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>

#include <iostream>
#include <vector>


//Loki
#include "loki/Typelist.h"
#include "loki/HierarchyGenerators.h"
#include "loki/TypeManip.h"
#include "loki/NullType.h"

typedef unsigned int uint;

//Core mechanism
#include "dynamicKernelTemplates.h"

//Define used to parametrize which numper of parameters
//to use in the sample kernel -> not used in the "library" part !
#define SAMPLE_NUMBER_OF_PARAMS			6  //2, 3 or 6
#define SAMPLE_NUMBER_OF_INTEGER_VALUES	2

///////////////////////////////////////
////User part, user defined sample ////
////kernels and associated functors////
///////////////////////////////////////

//A test kernel templated by 6 integer parameters
template <	int TParam1
		  ,	int TParam2
#if SAMPLE_NUMBER_OF_PARAMS >2
		  , int TParam3
#if SAMPLE_NUMBER_OF_PARAMS >3
		  , int TParam4
		  ,	int TParam5
		  ,	int TParam6 
#endif
#endif
						>
__global__ 
void myTemplatedKernel(uint *d_res){

	(*d_res)=TParam1+TParam2*10 
#if SAMPLE_NUMBER_OF_PARAMS >2
				+TParam3*100
#if SAMPLE_NUMBER_OF_PARAMS >3
				+TParam4*1000 +TParam5*10000+TParam6*100000
#endif
#endif
				;
}


struct MyKernelCallFunctor {
	///Kernel arguments are stored as members;
	uint *d_res;
	//Arguments are passed through the constructor
	MyKernelCallFunctor(uint *dres) : d_res(dres){}

	//Actual kernel call is made here.
	//TypeList is passed as a method template parameter.
	template<class TList>
	void call( TList tl){
		//Fetch template parameters from the typelist
		myTemplatedKernel <
					Loki::TL::TypeAt<TList, 0>::Result::value
				,	Loki::TL::TypeAt<TList, 1>::Result::value
#	if SAMPLE_NUMBER_OF_PARAMS >2
				,	Loki::TL::TypeAt<TList, 2>::Result::value
#	if SAMPLE_NUMBER_OF_PARAMS >3
				,	Loki::TL::TypeAt<TList, 3>::Result::value
				,	Loki::TL::TypeAt<TList, 4>::Result::value
				,	Loki::TL::TypeAt<TList, 5>::Result::value 
#	endif
#	endif
				>
			<<<32, 1>>>(d_res);	//Also pass function arguments
    }
};


//Dynamic template parameters emulation: integer with SAMPLE_NUMBER_OF_INTEGER_VALUES
void kernelCallEmul(int tv0, int tv1, 
#if SAMPLE_NUMBER_OF_PARAMS >2
					int tv2, 
#if SAMPLE_NUMBER_OF_PARAMS >3
					int tv3, int tv4, int tv5, 
#endif
#endif
					uint *d_res) {

	//Uses a Caller class
	//Second parameter is the number of "dynamic" parameters
	KernelCaller<Loki::NullType, SAMPLE_NUMBER_OF_PARAMS, MyKernelCallFunctor, SAMPLE_NUMBER_OF_INTEGER_VALUES-1>	//6 integer values between 0 and 9
		::call( &MyKernelCallFunctor(d_res),	//User call Functor.
				tv0, tv1
#if SAMPLE_NUMBER_OF_PARAMS >2
				, tv2
#if SAMPLE_NUMBER_OF_PARAMS >3
				, tv3, tv4, tv5
#endif
#endif
				);  //Dynamic template parameters
}

////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
int main( int argc, char** argv) {

	//Set cuda device
	cudaSetDevice(0);

	//Allocate test result
	uint *d_res;
	cudaMalloc(&d_res, 1*sizeof(uint));

	std::cout<<"Launch start...\n";

	//Call kernel call test function with some values
#if SAMPLE_NUMBER_OF_PARAMS==2
	kernelCallEmul(false, true, d_res);
#elif SAMPLE_NUMBER_OF_PARAMS==3
	kernelCallEmul(true, false, true, d_res);
#elif SAMPLE_NUMBER_OF_PARAMS==6
	kernelCallEmul(false, true, false, true, false, true, d_res);
#endif


	std::cout<<"Launch end\n";

	//Get back result value
	uint h_res;
	cudaMemcpy(&h_res, d_res, sizeof(uint), cudaMemcpyDeviceToHost);

	//Display result
	std::cout<<"Res: "<<h_res<<"\n";

	printf("\nPress ENTER to exit...\n");
    fflush( stdout);
    fflush( stderr);
    getchar();
}
